\ CONTENT       : Handles the object "Solution" and operations
\ LANGUAGE      : tForth
\ PROJECT       : hcc primes
\ CATEGORY      : EXAMPLE
\ AUTHOR        : AH 911108
\ LAST CHANGE   : AH 911119 FindMsBit rewritten, cosmetic
\ LAST CHANGE   : AH 911118 ?NonTrivial stack error
\ LAST CHANGE   : AH 911218 [ DEBUG ] if
\ LAST CHANGE   : MH 911216 Vnum has length MaxV+1
\ LAST CHANGE   : development

\ Assumes the module FACTORS.FRT has been loaded.


        REVISION -solution " Combining Solutions Version 1.00 "


DOC  Module 6
(*
          SOLUTIONS
          =========

      Solution ( "name" -- )
      creates a structure consisting of a bit array of
      MaxB cells, followed by two VNum's.
      This can best be in the form of a TO-object, such that
      %bits SomeSolution delivers the address of the bit-table
      %V1 SomeSolution delivers the address of the first VNum.
      %V2 SomeSolution delivers the address of the second VNum.
      There is are two scratch solutions SOL1// and SOL2//

      { For the curious :
        a Solution is a part of the puzzle that has to be fit together
        arrive at a divisor of TheNumber.
        It means that the produkt of all the primes represented by bits
        up in the bit array times V1^2 is equal to V2^2 modulo TheNumber.
        If we succeed in pasting solutions together such that all bits
        become zero: V1^2-V2^2 = n * TheNumber . Than
        gcd(TheNumber,V1-V2) is a divisor of TheNumber. }

      MakeSolution ( i -- flag )
      returns TRUE if a solution was found. Also this solution is put
      in SOL1// .
      Factorise (TheRoot+Offset+i)^2-TheNumber using only factors from
      primebase. If this succeeds, set a bit up in the bit-table of
      SOL1// for each prime that occurs with an odd power.
      Divide these primes out and put the square root of the
      remaining number in the first VNum of SOL1//
      The number (TheRoot+Offset+i) is put in the second VNum
      Example: TheNumber = 102 TheRoot = 10 Offset= 0 i = 2
      12^2-102=42 should be factorized. Result 2.3.7
      Put 11010000000... in the bit table.
      After dividing out 1 remains. Put this 1 in V1
      Put 10+0+2 = 12 in V2.
      Return flag "factoring has succeeded"
      N.B. factors < MaxPrime not present in PrimeBase are impossible.

      FindSolution// ( -- )
      after SieveAll// has been executed (module 5) look in the Sieve
      for all bytes greater than the number of bits in TheRoot.
      (So we cannot handle TheNumber >2^512 or more than 150 digits)
      The number represented can be factored using MakeSolution//
      and is sent to the central assembling task.

      SendSolution// ( -- )
      send solution (SOL1) to the central task.
      Append a signature such that the finder of this solution is
      known.
*)
ENDDOC

\ VS= compares the Vnumber with a single
: VS=           ( V,i -- flag )
   DUP IF
      SWAP @+ 1 = >R     \ Length
      @ =                \ Content
      R> AND
   ELSE
      DROP @ 0=
   ENDIF
;

        \ Message numbers, accepted by a Solutions TO-object
#20 =: %bits%  PRIVATE         \ get the address of the bit-table
#21 =: %V1%    PRIVATE         \ get the address of the first VNum.
#22 =: %V2%    PRIVATE         \ set the address of the second VNum.

        \ Message posters
: %bits         %bits%      %VAR !  ;
: %V1           %V1%        %VAR !  ;
: %V2           %V2%        %VAR !  ;

MaxB                    \ Bit-table
1+  MaxV +              \ First vnum
1+  MaxV + CELLS        \ Second vnum
                 CONSTANT #solution

\ Create the database structure for a Solution.
\ A TO-struct that accepts the messages above.
\ The content of a Solution S means :
\ Product over primes of the factorbase with bits up in %bits S
\ times %V1 S squared equals %V2 Squared.
\ Look at the output of PrintSolution for better understanding.

: Solution      CREATE                  \ "name" --
                #solution ALLOT
                DOES>                           \ <??> <pin> -- <?>
                   %VAR @  0 %VAR !     \ Reset to FROM
                   CASE
                     %addressof%  OF ( Leave address)       ENDOF
                     %sizeof%     OF DROP #solution         ENDOF
                     %bits%       OF  ( nothing)            ENDOF
                     %V1%         OF MaxB CELLS +           ENDOF
                     %V2%         OF MaxB MaxV 1+ + CELLS + ENDOF
                     DUP DEC. TRUE ABORT" Solution: unknown message"
                   ENDCASE ;

Solution SOL1
Solution SOL2
Solution SOL3


: SSmove  #solution CMOVE ;               \ <S1> <S2> -- <>

\ Auxiliary routine, of every two factors P in SS move one to VV
\ Leave TRUE if there remains a factor.
\ SS{new}/SS{old} = ( VV{new}/VV{old} )^2 * ( if flag then p else 1 )
: RemoveSquares ( SS VV P -- flag )
     LOCAL P
     LOCAL VV
     LOCAL SS
     FALSE              \ No odd factor
     BEGIN
        SS P VSMOD 0= WHILE
        IF              \ Already a factor
          SS P VS/MOD DROP
          VV P VS*
          FALSE
        ELSE
          SS P VS/MOD DROP
          TRUE
        ENDIF
     REPEAT
; PRIVATE

\ Store the double number D in a variable number V
: DV!   ( D,V -- )
     LOCAL V
     2 V !
     V 2 CELLS + !
     V 1 CELLS + !
     V COMPACT
;

\ Make solution tries to factor the number corresponding to the
\ offset D in the sieve. It uses only factors from the factor base.
\ Returns success or failure (too large factors)
\ On succes it leaves the solution in SOL1
\ The number gotten by multiplying all primes corresponding to bits
\ will be denoted by {bits}
\ There is some talk in the literature about the usefulnes of keeping
\ solutions where the remainder in SS is not 1.
: MakeSolution          ( D -- )
     %V2   SOL1 LOCAL V2
     %V1   SOL1 LOCAL V1
     %bits SOL1 LOCAL bits

     \ Initialize three fields of SOL1
     bits MaxB CELLS ERASE

     V1 1V!

     ( Offset, double ) V2 DV!
     V2 TheRoot VV+

     V2 V2 VV*
     PP TheNumber VV-
     PP SS VVmove

     \ Loop invariant: V1^2 * V2 * {bits}
     \ Starting with V1=1 and ending (hopefully) with V2=1
     \ In the loop we will try to move the square factors
     \ from V2 to V1 and the others to {bits}
     %primemark PrimeBase 1+ 1 DO      \ For all primes in primebase
        SS V1  { I %prime PrimeBase  }1
        {3 RemoveSquares }1 IF
           bits I SetBit
           SS 1 VS= IF LEAVE THEN
        THEN
     LOOP
\     CR V1 .VNUM
\     CR V2 .VNUM
\     CR SS .VNUM
     SS 1 VS=                   \ This condition means "fully factored"
;


\ Returns the index of the most significant bit up
\ Warning: Bit counting from 1 .. 32
\ 0 means no bits up. This marks the end of factoring ....
: FindMsBit     ( array[0..MaxB] -- I )
    FALSE LOCAL Found

    0 MaxB 1- DO
       DUP I CELLS + @ IF
            TRUE TO Found
            I CELLS + @ Scount-bits \ The bits in this cell
            I BITS/CELL * +  \ The number of bits in cells not inspected
            LEAVE
       THEN
   -1 +LOOP
   Found 0= IF
      DROP 0
   THEN
;


 \ Prints the solutions in some reasonable understandable format
 \ Do not confuse "solution" wit "factorization"
 \ A solution represents the mathematical identity printed here.
: PrintSolution          ( S -- )
     LOCAL bits
     bits MaxB CELLS + LOCAL V1
     V1 MaxV 1+ CELLS + LOCAL V2

     CR
     bits FindMsBit 1+ 1 ?DO
         bits I TestBit IF
             I %prime PrimeBase . ." * "
         THEN
     LOOP
     V1 .VNUM ." ^2 = " V2 .VNUM ." ^2 mod " TheNumber .VNUM
;

\ CombineSolution combines the solutions of SOL1 and SOL2
\ into SOL3. The highest primes present in the bit-table should
\ be the same, otherwise there is nothing to combine.
: CombineSolution   ( -- )
     %bits SOL1 FindMsBit LOCAL MsBit
\     MsBit %bits SOL2 FindMsBit  <> ABORT" No matching solutions"

     MsBit 0= ABORT" no bits in SOL1 in ComBineSolution"

     %V1 SOL1 %V1 SOL2 VV*
     PP TheNumber VV/MOD
     RR %V1 SOL3 VVmove

     %V2 SOL1 %V2 SOL2 VV*
     PP TheNumber VV/MOD
     RR %V2 SOL3 VVmove

     MaxB 0 DO
        %bits SOL1 I CELLS +  @
        %bits SOL2 I CELLS +  @    XOR
        %bits SOL3 I CELLS +  !
    LOOP

    MsBit 1+ 1 DO
        %bits SOL1 I TestBit IF
            %bits SOL2 I TestBit IF     \ Two bits cancel
                  %V1 SOL3
                  I %prime PrimeBase    \ Give a p^2 here
                  [ DEBUG ] [IF] DUP CR ." The Prime is " . [THEN]
                  VS*
                  %V1 SOL3 TheNumber VV/MOD
                  RR %V1 SOL3 VVmove
            THEN
        THEN
    LOOP
;

\ ?NonTrivial inspects the solution S and
\ returns whether it is a non trivial solution
: ?NonTrivial   ( S -- flag )
   DUP FindMsBit IF DROP TRUE EXIT THEN      \ Some of {bits} filled in
   MaxB CELLS +
   DUP MaxV 1+ CELLS +
   2DUP VV= IF
       2DROP FALSE                \ V1=V2
   ELSE
       SS VVmove SS SWAP VV+      \ Sum of numbers in SS
       SS TheNumber VV= 0=        \ V1 = -V2 modulo TheNumber
   THEN
;
